#ifndef __FRAGMENTEDMESSAGE_H__
#define __FRAGMENTEDMESSAGE_H__

#include<Arduino.h>

const int MESSAGE_FRAGMENT_SIZE = 31;

class FragmentedMessage
{
public:
  int NumFragments;
  char** Fragments;
  char* ChecksumFragment;

  FragmentedMessage()
  {
    NumFragments = 15;
    Fragments = new char*[NumFragments];
    for(int i =0; i < NumFragments; i++)
    {
      Fragments[i] = NULL;
    }

    ChecksumFragment = NULL;
  }

  FragmentedMessage(char* message, int senderID)
  {
    int numFragments = (strlen(message) / MESSAGE_FRAGMENT_SIZE) + 1;
    char** fragments = new char*[numFragments];
    memset(fragments, 0, numFragments);

    for(int fragmentIndex = 0; fragmentIndex < numFragments; fragmentIndex++)
    {
      char* fragment = new char[MESSAGE_FRAGMENT_SIZE +1];
      memset(fragment, 0, MESSAGE_FRAGMENT_SIZE +1);
      fragment[0] = (char)((fragmentIndex << 4) | (senderID % 16));

      strncpy((fragment + 1), (message + (MESSAGE_FRAGMENT_SIZE * fragmentIndex)), MESSAGE_FRAGMENT_SIZE);
      fragments[fragmentIndex] = fragment;
      //debugPrintMessageFragment(fragment);
    }

    char* checksumFragment = createChecksumFragment(message, senderID);

    //DebugPrintMessageFragment(checksumFragment);

    Fragments = fragments;
    NumFragments = numFragments;
    ChecksumFragment = checksumFragment;
  }

  ~FragmentedMessage()
  {
    for(int i = 0; i < NumFragments; i++)
    {
      if(Fragments[i] != NULL)
      {
        delete[] Fragments[i]; 
      }
    }

    if(ChecksumFragment != NULL)
    {
      delete[] ChecksumFragment;
    }
    
    if(Fragments != NULL)
    {
      delete[] Fragments;
    }
  }
  
  void DebugPrintMessageFragment(char* messageFragment)
  {
    //return;
    
     int senderID = DecodeFragmentSenderID(messageFragment);
     int fragmentNumber = DecodeFragmentNumber(messageFragment);
     
     Serial.print("Sender: ");
     Serial.println(senderID);
     Serial.print(" Fragment: ");
     Serial.println(fragmentNumber);
     

    Serial.print("[");

    for(int i = 1; i < MESSAGE_FRAGMENT_SIZE + 1; i++)
    {
      Serial.print((char)messageFragment[i]);
    }

    Serial.println("]");
    
    Serial.println("");
  }
  
  char* DefragmentMessage()
  {
    int activeFragments = countActiveFragments();

    //Serial.print("Active Fragments: ");
    //Serial.println(activeFragments);

    char* messageBuffer = new char[MESSAGE_FRAGMENT_SIZE * activeFragments];
    memset(messageBuffer, 0, MESSAGE_FRAGMENT_SIZE * activeFragments);

    for(int fragmentIndex = 0; fragmentIndex < activeFragments; fragmentIndex++)
    {
      if(Fragments[fragmentIndex] == NULL)
      {
        continue;
      }

      //DebugPrintMessageFragment(Fragments[fragmentIndex]);

      strncpy((messageBuffer + (MESSAGE_FRAGMENT_SIZE * fragmentIndex)), (Fragments[fragmentIndex] + 1), MESSAGE_FRAGMENT_SIZE);
      //Serial.println(messageBuffer);
    }

    //Serial.print("Defragmented: {");
    //Serial.print(messageBuffer);
    //Serial.println("}");
    
    int bufferChecksum = calculateChecksum(messageBuffer);
    int recvdMessageChecksum = getChecksumFromFragmentedMessage();

    //Serial.print("buffer chk: ");
    //Serial.print(bufferChecksum);
    //Serial.print(" recvd chk: ");
    //Serial.println(recvdMessageChecksum);
      
    if(bufferChecksum != recvdMessageChecksum)
    {
      delete[] messageBuffer;
      return NULL;
    }
    
    return messageBuffer;
  }
  
  int HasChecksumFragment()
  {
    return ChecksumFragment != NULL;
  }
  
  void AppendMessageFragment(char* messageFragment)
  {
    char* fragmentCopy = new char[MESSAGE_FRAGMENT_SIZE + 1];
    memcpy(fragmentCopy, messageFragment, MESSAGE_FRAGMENT_SIZE +1);

    int fragmentNumber = DecodeFragmentNumber(fragmentCopy);

    if(fragmentNumber == 0x0F)
    {
      if(ChecksumFragment != NULL)
      {
        delete[] ChecksumFragment;
      }
      ChecksumFragment = fragmentCopy; 
    }
    else    
    {
      if(Fragments[fragmentNumber] != NULL)
      {
        delete[] Fragments[fragmentNumber];
      }
      Fragments[fragmentNumber] = fragmentCopy; 
    }
  }
  
  static int DecodeFragmentSenderID(char* message)
  {
    return 0x0F & message[0];
  }
  
  static int DecodeFragmentNumber(char* message)
  {
    return (message[0] >> 4) & 0x0F;
  }

private:
  char* createChecksumFragment(char* message, int senderID)
  {
    char* fragment = new char[MESSAGE_FRAGMENT_SIZE +1];
    memset(fragment, 0, MESSAGE_FRAGMENT_SIZE +1);
    fragment[0] = (char)(0xF0 | (senderID % 16));

    int checksum = calculateChecksum(message);

    snprintf(fragment + 1, MESSAGE_FRAGMENT_SIZE, "%d", checksum);

    return fragment;
  }

  int calculateChecksum(char* message)
  {
    char checksum = 0;
    for(int i = 0; i < strlen(message); i++)
    {
      checksum ^= message[i];
    }

    return checksum;
  }
  
  int getChecksumFromFragmentedMessage()
  {
    int checksum = atoi(ChecksumFragment + 1);
    return checksum;
  }
  
  int countActiveFragments()
  {
    int activeFragments = 0;
    for(; (activeFragments < NumFragments) && (Fragments[activeFragments] != NULL); activeFragments ++) {}
    return activeFragments;
  }
};

#endif

